@@ -2,13 +2,16 @@
 
 import ctypes
 from ctypes import byref, c_int, c_void_p
+import logging
 from pathlib import Path
 import typing
-from typing import Dict, List, Optional, Union
+from typing import Dict, List, Mapping, Optional, Tuple, Union
 
 import numpy as np
 import yaml
 
+from furiosa import registry
+
 from . import envs
 from ._api import LIBNUX
 from ._api.v1 import convert_to_cchar_array, decref, increase_ref_count, runtime_version
@@ -69,7 +72,7 @@ def _create_session_options(
     worker_num: Optional[int] = None,
     batch_size: Optional[int] = None,
     compiler_hints: Optional[bool] = None,
-    compile_config: Optional[Dict[str, object]] = None,
+    compiler_config: Optional[Mapping[str, object]] = None,
     compiler_log: Optional[Path] = None,
     input_queue_size: Optional[int] = None,
     output_queue_size: Optional[int] = None,
@@ -83,9 +86,9 @@ def _create_session_options(
         LIBNUX.nux_session_option_set_batch_size(options, batch_size)
     if compiler_hints:
         LIBNUX.nux_session_option_enable_compiler_hints(options, compiler_hints)
-    if compile_config:
-        compile_config = yaml.dump(compile_config).encode()
-        err = LIBNUX.nux_session_option_set_compiler_config(options, compile_config)
+    if compiler_config:
+        compiler_config = yaml.dump(compiler_config).encode()
+        err = LIBNUX.nux_session_option_set_compiler_config(options, compiler_config)
         if is_err(err):
             raise into_exception(err)
     if compiler_log:
@@ -109,7 +112,7 @@ class Session(Model):
         worker_num: Optional[int] = None,
         batch_size: Optional[int] = None,
         compiler_hints: bool = True,
-        compile_config: Optional[Dict[str, object]] = None,
+        compiler_config: Optional[Mapping[str, object]] = None,
     ):
         profiler_path = envs.profiler_output()
         if profiler_path is not None:
@@ -126,7 +129,7 @@ class Session(Model):
             worker_num=worker_num,
             batch_size=batch_size,
             compiler_hints=compiler_hints,
-            compile_config=compile_config,
+            compiler_config=compiler_config,
             compiler_log=log_path,
         )
         model_image = _model_image(model)
@@ -178,9 +181,7 @@ class Session(Model):
 
         return outputs
 
-    def run_with(
-        self, outputs: typing.List[str], inputs: Dict[str, Union[np.ndarray]]
-    ) -> TensorArray:
+    def run_with(self, outputs: typing.List[str], inputs: Dict[str, np.ndarray]) -> TensorArray:
         """
         Runs an inference task with `inputs`
 
@@ -221,10 +222,21 @@ class Session(Model):
 
         return output_tensors
 
-    def close(self):
-        """Close the session and release all resources belonging to the session"""
+    # _LIBNUX is meant to keep a valid reference to the Nux library at
+    # interpreter shutdown. From https://bugs.python.org/issue5099#msg80855:
+    #
+    # > At interpreter shutdown, the module's global variables are set to None
+    # > before the module itself is released. __del__ methods may be called in
+    # > those precarious circumstances, and should not rely on any global
+    # > state.
+    def close(self, *, _LIBNUX=LIBNUX):
+        """Close the session and release all resources belonging to the session
+
+        _LIBNUX is only for internal use and not supposed to be specified by a
+        user.
+        """
         if hasattr(self, "ref") and self.ref:
-            LIBNUX.nux_session_destroy(self.ref)
+            _LIBNUX.nux_session_destroy(self.ref)
             self.ref = None
 
         if hasattr(self, "profiler") and self.profiler:
@@ -249,7 +261,7 @@ class CompletionQueue:
         self,
         ref: c_void_p,
         context_ty: Optional[type],
-        output_descs: [TensorDesc],
+        output_descs: List[TensorDesc],
         profiler,
         profiler_file,
     ):
@@ -261,7 +273,7 @@ class CompletionQueue:
         self.profiler = profiler
         self.profiler_file = profiler_file
 
-    def recv(self, timeout: Optional[int] = None) -> (object, TensorArray):
+    def recv(self, timeout: Optional[int] = None) -> Tuple[object, TensorArray]:
         """Receives the prediction results which are asynchronously coming from AsyncSession
 
         If there are already prediction outputs, it will return immediately.
@@ -290,7 +302,7 @@ class CompletionQueue:
         context_ref = ctypes.py_object(None)
         outputs_ref = c_void_p(None)
 
-        if timeout:
+        if timeout is not None:
             if timeout < 0:
                 raise RuntimeError("the timeout duration must be a positive integer")
             self.queue_ok = LIBNUX.nux_completion_queue_next_timeout(
@@ -304,14 +316,24 @@ class CompletionQueue:
         decref(context_ref)
 
         if is_ok(err):
-            return context_val, TensorArray(outputs_ref, self.output_descs, allocated=False)
+            return context_val, TensorArray(outputs_ref, self.output_descs, allocated=True)
 
         raise into_exception(err)
 
-    def close(self):
+    # _LIBNUX is meant to keep a valid reference to the Nux library at
+    # interpreter shutdown. From https://bugs.python.org/issue5099#msg80855:
+    #
+    # > At interpreter shutdown, the module's global variables are set to None
+    # > before the module itself is released. __del__ methods may be called in
+    # > those precarious circumstances, and should not rely on any global
+    # > state.
+    def close(self, *, _LIBNUX=LIBNUX):
         """Closes this completion queue.
 
         If it is closed, AsyncSession also will stop working.
+
+        _LIBNUX is only for internal use and not supposed to be specified by a
+        user.
         """
         if hasattr(self, "profiler") and self.profiler:
             self.profiler.__exit__(None, None, None)
@@ -319,7 +341,7 @@ class CompletionQueue:
             self.profiler = None
 
         if hasattr(self, "ref") and self.ref:
-            LIBNUX.nux_completion_queue_destroy(self.ref)
+            _LIBNUX.nux_completion_queue_destroy(self.ref)
             self.ref = None
 
     def __enter__(self):
@@ -349,8 +371,6 @@ class AsyncSession(Model):
         self._as_parameter_ = self._ref
         super().__init__()
 
-        self._tensor_buffers = self.allocate_inputs()
-
     def _get_model_ref(self) -> c_void_p:
         if self._ref:
             return LIBNUX.nux_async_session_get_model(self)
@@ -370,22 +390,33 @@ class AsyncSession(Model):
             values: Input values
             context: an additional context to identify the prediction request
         """
-        _fill_all_tensors(values, self._tensor_buffers)
+        _inputs = self.allocate_inputs()
+        _inputs = _fill_all_tensors(values, _inputs)
         # manually increase reference count to keep the context object while running
         increase_ref_count(context)
-        err = LIBNUX.nux_async_session_run(self._ref, context, self._tensor_buffers)
+        err = LIBNUX.nux_async_session_run(self._ref, context, _inputs)
 
         if is_err(err):
             raise into_exception(err)
 
-    def close(self):
+    # _LIBNUX is meant to keep a valid reference to the Nux library at
+    # interpreter shutdown. From https://bugs.python.org/issue5099#msg80855:
+    #
+    # > At interpreter shutdown, the module's global variables are set to None
+    # > before the module itself is released. __del__ methods may be called in
+    # > those precarious circumstances, and should not rely on any global
+    # > state.
+    def close(self, *, _LIBNUX=LIBNUX):
         """Closes this session
 
         After a session is closed, CompletionQueue will return an error
         if CompletionQueue.recv() is called.
+
+        _LIBNUX is only for internal use and not supposed to be specified by a
+        user.
         """
         if self._ref:
-            LIBNUX.nux_async_session_destroy(self._ref)
+            _LIBNUX.nux_async_session_destroy(self._ref)
             self._ref = None
 
     def __enter__(self):
@@ -399,40 +430,58 @@ class AsyncSession(Model):
 
 
 def create(
-    model: Union[bytes, str, Path],
+    model: Union[bytes, str, Path, registry.Model],
     device: str = None,
     worker_num: int = None,
     batch_size: int = None,
-    compile_config: Dict[str, object] = None,
+    compiler_config: Mapping[str, object] = None,
     compiler_hints: bool = True,
 ) -> Session:
     """Creates a session for a model
 
     Args:
-        model (bytes, str, Path): a byte string containing a model image or \
-        a path string of a model image file
+        model (bytes, str, Path, Model): a byte string containing a model image or \
+        a path string of a model image file or `furiosa.registry.Model`
         device: NPU device (str) (e.g., npu0pe0, npu0pe0-1)
         worker_num: Number of workers
         batch_size: Batch size of input tensors
-        compile_config (Dict[str, object]): Compile config
+        compiler_config (Mapping[str, object]): Compile config
         compiler_hints: Print compiler hints if True (default: True)
 
     Returns:
         the session for a given model, allowing to run predictions. \
         Session is a thread safe.
     """
+    if isinstance(model, registry.Model):
+        use_fusion = (device or envs.current_npu_device()).endswith("pe0-1")
+        use_enf = model.enf is not None and use_fusion
+
+        if compiler_config is not None and model.compiler_config is not None:
+            logging.warning(
+                "Model's compiler config is ignored because an explicit compiler config is passed to session.create()"
+            )
+
+        return Session(
+            model=model.enf if use_enf else model.source,
+            device=device,
+            worker_num=worker_num,
+            batch_size=batch_size,
+            compiler_config=compiler_config or model.compiler_config,
+            compiler_hints=compiler_hints,
+        )
+
     return Session(
         model=model,
         device=device,
         worker_num=worker_num,
         batch_size=batch_size,
-        compile_config=compile_config,
+        compiler_config=compiler_config,
         compiler_hints=compiler_hints,
     )
 
 
 def create_async(
-    model: Union[bytes, str, Path],
+    model: Union[bytes, str, Path, registry.Model],
     context_ty: Optional[type] = None,
     device: Optional[str] = None,
     worker_num: Optional[int] = None,
@@ -440,13 +489,13 @@ def create_async(
     compiler_hints: Optional[bool] = True,
     input_queue_size: Optional[int] = None,
     output_queue_size: Optional[int] = None,
-    compile_config: Optional[Dict[str, object]] = None,
-) -> (AsyncSession, CompletionQueue):
+    compiler_config: Optional[Mapping[str, object]] = None,
+) -> Tuple[AsyncSession, CompletionQueue]:
     """Creates a pair of the asynchronous session and the completion queue for a given model
 
     Args:
-        model (bytes, str, Path): a byte string containing a model image or \
-        a path string of a model image file
+        model (bytes, str, Path, Model): a byte string containing a model image or \
+        a path string of a model image file or `furiosa.registry.Model`
         context_ty (type): Type for passing context from AsyncSession to CompletionQueue
         device: NPU device (str) (e.g., npu0pe0, npu0pe0-1)
         worker_num: Number of workers
@@ -454,7 +503,7 @@ def create_async(
         compiler_hints: Print compiler hints if True (default: True)
         input_queue_size: The input queue size, and it must be > 0 and < 2^31.
         output_queue_size: The output queue size, and it must be be > 0 and < 2^31.
-        compile_config (Dict[str, object]): Compile config
+        compiler_config (Mapping[str, object]): Compile config
 
     Returns:
         A pair of the asynchronous session and the completion queue. \
@@ -462,6 +511,26 @@ def create_async(
         the completion queue allows users to receive the prediction outputs \
         asynchronously.
     """
+    if isinstance(model, registry.Model):
+        use_fusion = (device or envs.current_npu_device()).endswith("pe0-1")
+        use_enf = model.enf is not None and use_fusion
+
+        if compiler_config is not None and model.compiler_config is not None:
+            logging.warning(
+                "Model's compiler config is ignored because an explicit compiler config is passed to session.create_async()"
+            )
+
+        return create_async(
+            model=model.enf if use_enf else model.source,
+            context_ty=context_ty,
+            device=device,
+            worker_num=worker_num,
+            batch_size=batch_size,
+            compiler_hints=compiler_hints,
+            input_queue_size=input_queue_size,
+            output_queue_size=output_queue_size,
+            compiler_config=compiler_config or model.compiler_config,
+        )
 
     try:
         if device is None:
@@ -483,7 +552,7 @@ def create_async(
             device=device,
             worker_num=worker_num,
             batch_size=batch_size,
-            compile_config=compile_config,
+            compiler_config=compiler_config,
             compiler_hints=compiler_hints,
             compiler_log=log_path,
             input_queue_size=input_queue_size,
